#include "ch.h"
#include "hal.h"

#include "lpc1114_ump.h"
#include "lpc1114_registers.h"
#include "console_input.h"

// Version 0.1 (works until version 255.255).
#define FIRMWARE_VERSION_MSB (0)
#define FIRMWARE_VERSION_LSB (1)

/*
                         LPC1114FN28
                        +-----U-----+
                P0.8  ##| 1      28 |## P0.7
                P0.9  ##| 2      27 |## P0.4
         SWCLK  P0.10 ##| 3      26 |## P0.3
         R      P0.11 ##| 4      25 |## P0.2
                P0.5  ##| 5      24 |## P0.1  (ISP)
                P0.6  ##| 6      23 |## P0.0  RESET
                VDDA  ##| 7      22 |## VSS
                VSSA  ##| 8      21 |## VDD
         R      P1.0  ##| 9      20 |## XTALI
         R      P1.1  ##| 10     19 |## XTALO
         R      P1.2  ##| 11     18 |## P1.9
         SWDIO  P1.3  ##| 12     17 |## P1.8
                P1.4  ##| 13     16 |## P1.7  RXD
                P1.5  ##| 14     15 |## P1.6  RXD
                        +-----------+
*/

/*
 * Configure pinstrapping pins as inputs with pull-ups enabled,
 * read their values and return them after restoring their original configuration.
 * The Pinstrapping pins used are PIO0_2, PIO0_3, PIO0_8 and PIO0_9.
 * These are used as outputs in this design.
 */
static int readPinStapping(void)
{
    /*
     *  COL1: PIO0_3: ID0
     *  COL2: PIO0_8: ID1
     *  COL3: PIO0_9: ID2
     *  COL4: PIO0_2: ID3
     */

    const uint32_t iocon[4] = { LPC_IOCON->PIO0_3, LPC_IOCON->PIO0_8, LPC_IOCON->PIO0_9, LPC_IOCON->PIO0_2 };

    //-- setup column selector (outputs)
    palSetGroupMode(GPIO0, PAL_PORT_BIT(3) | PAL_PORT_BIT(8) | PAL_PORT_BIT(9) | PAL_PORT_BIT(2) | PAL_PORT_BIT(0),
		    0, PAL_MODE_INPUT);
    LPC_IOCON->PIO0_3  = 0xC0 | (2<<3) | 0; // COL1 - Digital function GPIO, pull-up
    LPC_IOCON->PIO0_8  = 0xC0 | (2<<3) | 0; // COL2 - Digital function GPIO, pull-up
    LPC_IOCON->PIO0_9  = 0xC0 | (2<<3) | 0; // COL3 - Digital function GPIO, pull-up
    LPC_IOCON->PIO0_2  = 0xC0 | (2<<3) | 0; // COL4 - Digital function GPIO, pull-up

    uint8_t pinstrap =
	((palReadPad(GPIO0, 3) == 0) ? 1 : 0) +
	((palReadPad(GPIO0, 8) == 0) ? 2 : 0) +
	((palReadPad(GPIO0, 9) == 0) ? 4 : 0) +
	((palReadPad(GPIO0, 2) == 0) ? 8 : 0);

    LPC_IOCON->PIO0_3  = iocon[0];
    LPC_IOCON->PIO0_8  = iocon[1];
    LPC_IOCON->PIO0_9  = iocon[2];
    LPC_IOCON->PIO0_2  = iocon[3];

    // Restore GPIO Settings

    return pinstrap;
}

static Mailbox * g_inputEventMailbox = 0;

static void frameReceivedCallback(struct UmpFrame *_frame)
{
#ifndef DEBUG_AQUIRE_CHANNEL
    // reply to received frames. Put frame into rx-queue.xc

    if ((_frame->size >= 6) && (_frame->data[0] == 'L'))
    {
	setLedValues(_frame->data[1], _frame->data[2], _frame->data[3], (_frame->data[4]<<8) + _frame->data[5]);
    }
    else
    {
#define REQUEST_MESSAGE(sender) ((15<<24) | (sender))
	const msg_t m = REQUEST_MESSAGE(_frame->source);
	chMBPost (g_inputEventMailbox, m, TIME_INFINITE);
    }
#endif
}


#ifndef DEBUG_AQUIRE_CHANNEL
struct UmpFrame eventFrame;

static void flushEventFrame(const uint8_t _destination)
{
    while (!lpc1114_ump_transmitter_ready())
    {
	chThdSleepMilliseconds(5);
    }
    eventFrame.dest = _destination;
    lpc1114_ump_send_frame(&eventFrame);
    while (!lpc1114_ump_transmitter_ready())
    {
	chThdSleepMilliseconds(5);
    }
    eventFrame.size = 0;
}
#endif

/*
 * Application entry point.
 */
int main(void)
{
    /*
     * System initializations.
     * - HAL initialization, this also initializes the configured device drivers
     *   and performs the board-specific initializations.
     * - Kernel initialization, the main() function becomes a thread and the
     *   RTOS is active.
     */
    halInit();
    chSysInit();

  /*
   * create a maylbox for all the input events (fader, keyboard, encoder,...)
   */
#define InputEventMailboxSize 32
    msg_t inputEventMailboxBuffer[InputEventMailboxSize];
    Mailbox inputEventMailbox;
    chMBInit (&inputEventMailbox, inputEventMailboxBuffer, InputEventMailboxSize);
    g_inputEventMailbox = &inputEventMailbox;

    //-- Startup UPM layer
    lpc1114_ump_startup(frameReceivedCallback);
    lpc1114_ump_set_my_id(0x10 + readPinStapping());

    //--startup console input
    initConsoleInput(&inputEventMailbox);

#ifdef DEBUG_AQUIRE_CHANNEL
    while (1)
    {
	chThdSleepMilliseconds(1000);
    }
#else
    //-- prefill fields in the event frame.
    eventFrame.source = lpc1114_ump_my_id();
    eventFrame.flags = 1;
    eventFrame.size = 0;

    // ID=0 means Selftest
    if (eventFrame.source == 0x10)
    {
	setLedValues(0, 0, 1, 0);
	setLedValues(0, 0, 2, 300);
	setLedValues(0, 0, 4, 600);
	setLedValues(0, 0, 8, 1000);

	while (TRUE)
	{
	    msg_t msg;
	    switch (chMBFetch (&inputEventMailbox, &msg, MS2ST(10) /*TIME_INFINITE*/ /*TIME_IMMEDIATE*/))
	    {
	    case RDY_OK:
		switch ((unsigned char)(msg>>24))
		{
		case 1: // Fader  { 8'd1, 8'd<channel>, 16'd<value> }
	            {
			int value = (unsigned short)(msg&0xFFFF);
			const unsigned char fader = (unsigned char)((msg>>16)&0xFF);
			if (value < 0)
			    value = 0;
			else if (value > 1000)
			    value = 1000;
			setLedValues(fader, 3, 15, value);
			setLedValues(fader, 4, 15, 1000-value);
		    }
		    break;
		
		case 2: // Fader Button
	            {
			const unsigned char fader = (unsigned char)((msg>>16)&0xFF);
			const unsigned short keycode = (unsigned short)(msg&0x7FFF);
			const char pressed = (msg>>15)&1;

			setLedValues(fader, keycode, 15, pressed ? 1000 : 0);
		    }
		    break;

		case 14: // Heardbeat from console
		    break;

		default:
		    break;
		}
		break;
		
	    case RDY_RESET:
		break;
		
	    case RDY_TIMEOUT:
		break;
	    }
	}
    }
    else // --Normal operation
    {
	uint8_t event_receiver_id = 0xff;

	while (TRUE)
	{
	    msg_t msg;
	    switch (chMBFetch (&inputEventMailbox, &msg, MS2ST(10) /*TIME_INFINITE*/ /*TIME_IMMEDIATE*/))
	    {
	    case RDY_OK:
		switch ((unsigned char)(msg>>24))
		{
		case 1: // Fader  { 8'd1, 8'd<channel>, 16'd<value> }
		    if (event_receiver_id != 0xff)
		    {
			const int value = (unsigned short)(msg&0xFFFF);
			const unsigned char fader = (unsigned char)((msg>>16)&0xFF);

			eventFrame.data[eventFrame.size+0] = (uint8_t)'F';
			eventFrame.data[eventFrame.size+1] = (uint8_t)fader;
			eventFrame.data[eventFrame.size+2] = (uint8_t)(value >> 8);
			eventFrame.data[eventFrame.size+3] = (uint8_t)value;
			eventFrame.size += 4;
		    }
		    break;

		case 2: // Fader Button
		    if (event_receiver_id != 0xff)
		    {
			const unsigned char fader = (unsigned char)((msg>>16)&0xFF);
			const unsigned short keycode = (unsigned short)(msg&0x7FFF);
			const char pressed = (msg>>15)&1;
		    
			eventFrame.data[eventFrame.size+0] = (uint8_t)'B';
			eventFrame.data[eventFrame.size+1] = (uint8_t)fader;
			eventFrame.data[eventFrame.size+2] = (uint8_t)keycode;
			eventFrame.data[eventFrame.size+3] = (uint8_t)pressed;
			eventFrame.size += 4;
		    }
		    break;

		case 4: // Console Key
		    if (event_receiver_id != 0xff)
		    {
			const unsigned short keycode = (unsigned short)msg;
			const char pressed = (msg>>23)&1;
			(void)keycode;
			(void)pressed;
		    }
		    break;

		case 14: // Heardbeat from console
		    break;

		case 15: // Request for messages
	            {
			event_receiver_id = (msg&0xff);

			if (eventFrame.size > 0)
			{
			    if (lpc1114_ump_transmitter_ready())
			    {
				flushEventFrame(event_receiver_id);
			    }
			}

			eventFrame.dest = event_receiver_id;
			eventFrame.size = 3;
			eventFrame.data[0] = 'A'; // Ack
			eventFrame.data[1] = FIRMWARE_VERSION_MSB; // Firmware Version MSB
			eventFrame.data[2] = FIRMWARE_VERSION_LSB; // Firmware Version LSB
			lpc1114_ump_send_frame(&eventFrame);
			while (!lpc1114_ump_transmitter_ready())
			{
			    chThdSleepMilliseconds(5);
			}
			eventFrame.size = 0;
		    }
		    break;

		default:
		    break;
		}

		if (eventFrame.size >= 64)
		{
		    flushEventFrame(event_receiver_id);
		}
		break;

	    case RDY_RESET:
		break;

	    case RDY_TIMEOUT:
		if (eventFrame.size > 0)
		{
		    if (lpc1114_ump_transmitter_ready())
		    {
			flushEventFrame(event_receiver_id);
		    }
		}
		break;
	    }
	}
    }
#endif
    lpc1114_ump_shutdown();

    return 0;
}

